"""
The package that provides generating methods for all parts
of the pdelab driver.

Currently, these are hardcoded as strings. It would be possible
to switch these to cgen expression. OTOH, there is not much to be
gained there.

NB: Previously this __init__.py was a module driver.py. As it was growing,
    we made it a package. Some content could and should be separated into
    new modules within this package!
"""
from dune.perftool.error import PerftoolCodegenError
from dune.perftool.generation import (generator_factory,
                                      get_global_context_value,
                                      global_context,
                                      include_file,
                                      cached,
                                      pre_include,
                                      preamble,
                                      )
from dune.perftool.options import get_option

# Have a global variable with the entire form data. This allows functions that depend
# deterministically on the entire data set to directly access it instead of passing it
# through the entire generator chain.
_driver_data = {}


# Have a function access this global data structure
def set_driver_data(formdatas, data):
    assert (len(formdatas) <= 2)
    if len(formdatas) == 1:
        _driver_data['form'] = formdatas[0].preprocessed_form
        _driver_data['formdata'] = formdatas[0]
    else:
        mass_index = mass_form_index(formdatas, data)
        if mass_index is None:
            raise NotImplementedError("Form for mass matrix needs to have name 'mass' in ufl file.")
        _driver_data['mass_form'] = formdatas[mass_index].preprocessed_form
        _driver_data['mass_formdata'] = formdatas[mass_index]
        _driver_data['form'] = formdatas[1 - mass_index].preprocessed_form
        _driver_data['formdata'] = formdatas[1 - mass_index]

    _driver_data['data'] = data


def get_dimension():
    return _driver_data['form'].ufl_cell().geometric_dimension()


def get_cell():
    return _driver_data['form'].ufl_cell().cellname()


def get_test_element():
    return _driver_data['form'].arguments()[0].ufl_element()


def get_trial_element():
    return _driver_data['form'].coefficients()[0].ufl_element()


def get_formdata():
    return _driver_data['formdata']


def get_mass_formdata():
    return _driver_data["mass_formdata"]


def is_stationary():
    return 'mass_form' not in _driver_data


def form_name_suffix(name, formdata):
    from dune.perftool.pdelab.localoperator import name_form
    data = get_global_context_value('data')
    form_name = name_form(formdata, data)
    return name + '_' + form_name


def get_object(name):
    return _driver_data['data'].object_by_name.get(name, None)


def mass_form_index(formdatas, data):
    for index, formdata in enumerate(formdatas):
        try:
            if data.object_names[id(formdata.original_form)] == 'mass':
                return index
        except KeyError:
            continue


def is_linear(form=None):
    '''Test if form is linear in trial function'''
    if form is None:
        form = get_formdata().original_form
    from ufl import derivative
    from ufl.algorithms import expand_derivatives
    jacform = expand_derivatives(derivative(form, form.coefficients()[0]))
    for coeff in jacform.coefficients():
        if 0 == coeff.count():
            return False

    return True


def isLagrange(fem):
    return fem._short_name is 'CG'


def isSimplical(cell):
    # Cells can be identified through strings *or* ufl objects
    if not isinstance(cell, str):
        cell = cell.cellname()
    return cell in ["vertex", "interval", "triangle", "tetrahedron"]


def isQuadrilateral(cell):
    # Cells can be identified through strings *or* ufl objects
    if not isinstance(cell, str):
        cell = cell.cellname()
    return cell in ["vertex", "interval", "quadrilateral", "hexahedron"]


def isPk(fem):
    return isLagrange(fem) and isSimplical(fem.cell())


def isQk(fem):
    return isLagrange(fem) and isQuadrilateral(fem.cell())


def isDG(fem):
    return fem._short_name is 'DG'


def FEM_name_mangling(fem):
    from ufl import MixedElement, VectorElement, FiniteElement, TensorElement, TensorProductElement
    if isinstance(fem, VectorElement):
        return FEM_name_mangling(fem.sub_elements()[0]) + "_" + str(fem.num_sub_elements())
    if isinstance(fem, TensorElement):
        return FEM_name_mangling(fem.sub_elements()[0]) + "_" + "_".join(str(i) for i in fem.value_shape())
    if isinstance(fem, MixedElement):
        name = ""
        for elem in fem.sub_elements():
            if name is not "":
                name = name + "_"
            name = name + FEM_name_mangling(elem)
        return name
    if isinstance(fem, FiniteElement):
        if isPk(fem):
            return "P" + str(fem._degree)
        if isQk(fem):
            return "Q" + str(fem._degree)
        if isDG(fem):
            return "DG" + str(fem._degree)
    if isinstance(fem, TensorProductElement):
        assert(len(set(subel._short_name for subel in fem.sub_elements())) == 1)

        if isLagrange(fem.sub_elements()[0]):
            return "TensorQ" + '_'.join(map(str, fem._degree))
        if isDG(fem.sub_elements()[0]):
            return "TensorDG" + '_'.join(map(str, fem._degree))
        raise NotImplementedError("fem name mangling")

    raise NotImplementedError("FEM NAME MANGLING")


def _flatten_list(l):
    if isinstance(l, (tuple, list)):
        for i in l:
            for ni in _flatten_list(i):
                yield ni
    else:
        yield l


def _unroll_list_tensors(expr):
    from ufl.classes import ListTensor
    if isinstance(expr, ListTensor):
        for op in expr.ufl_operands:
            yield op
    else:
        yield expr


def unroll_list_tensors(data):
    for expr in data:
        for e in _unroll_list_tensors(expr):
            yield e


def preprocess_leaf_data(element, data, applyZeroDefault=True):
    data = get_object(data)
    if data is None and not applyZeroDefault:
        return None

    from ufl import MixedElement
    if isinstance(element, MixedElement):
        # data is None -> use 0 default
        if data is None:
            data = (0,) * element.value_size()

        # Flatten nested lists
        data = tuple(i for i in _flatten_list(data))

        # Expand any list tensors
        data = tuple(i for i in unroll_list_tensors(data))

        assert len(data) == element.value_size()
        return data
    else:
        # Do not return lists for non-MixedElement
        if not isinstance(data, (tuple, list)):
            return (data,)
        else:
            assert len(data) == 1
            return data


def name_inifile():
    # TODO pass some other option here.
    return "argv[1]"


@preamble
def parse_initree(varname):
    include_file("dune/common/parametertree.hh", filetag="driver")
    include_file("dune/common/parametertreeparser.hh", filetag="driver")
    filename = name_inifile()
    return ["Dune::ParameterTree initree;", "Dune::ParameterTreeParser::readINITree({}, {});".format(filename, varname)]


def name_initree():
    parse_initree("initree")
    # TODO we can get some other ini file here.
    return "initree"


@preamble
def define_mpihelper(name):
    include_file("dune/common/parallel/mpihelper.hh", filetag="driver")
    return "Dune::MPIHelper& {} = Dune::MPIHelper::instance(argc, argv);".format(name)


def name_mpihelper():
    name = "mpihelper"
    define_mpihelper(name)
    return name


@preamble
def check_parallel_execution():
    from dune.perftool.pdelab.driver.gridfunctionspace import name_leafview
    gv = name_leafview()
    return ["if ({}.comm().size()==1){{".format(gv),
            '  std::cout << "This program should be run in parallel!"  << std::endl;',
            "  return 1;",
            "}"]


def generate_driver(formdatas, data):
    # Add check to c++ file if this program should only be used in parallel mode
    if get_option("parallel"):
        check_parallel_execution()

    # Entrypoint for driver generation
    if get_option("opcounter") or get_option("time_opcounter"):
        if get_option("time_opcounter"):
            assert(not get_option("opcounter"))
        assert(any(_driver_data['form'].ufl_cell().cellname() in x for x in
                   ["vertex", "interval", "quadrilateral", "hexahedron"]))
        # In case of operator conunting we only assemble the matrix and evaluate the residual
        # assemble_matrix_timer()
        from dune.perftool.pdelab.driver.timings import apply_jacobian_timer, evaluate_residual_timer
        from dune.perftool.loopy.target import type_floatingpoint
        pre_include("#define HP_TIMER_OPCOUNTER {}".format(type_floatingpoint()), filetag="driver")
        evaluate_residual_timer()
        apply_jacobian_timer()
    elif is_stationary():
        from dune.perftool.pdelab.driver.solve import dune_solve
        vec = dune_solve()
        from dune.perftool.pdelab.driver.vtk import vtkoutput
        vtkoutput()
    else:
        from dune.perftool.pdelab.driver.instationary import solve_instationary
        solve_instationary()

    from dune.perftool.pdelab.driver.error import compare_L2_squared
    if get_option("compare_l2errorsquared"):
        compare_L2_squared()

    # Make sure that timestream is declared before retrieving chache items
    if get_option("instrumentation_level") >= 1:
        from dune.perftool.pdelab.driver.timings import setup_timer, name_timing_stream
        setup_timer()
        timestream = name_timing_stream()

    from dune.perftool.pdelab.driver.error import return_statement
    return_statement()

    from dune.perftool.generation import retrieve_cache_items
    from cgen import FunctionDeclaration, FunctionBody, Block, Value
    driver_signature = FunctionDeclaration(Value('bool', 'driver'), [Value('int', 'argc'), Value('char**', 'argv')])
    contents = [i for i in retrieve_cache_items("preamble", make_generable=True)]

    from cgen import Line
    if get_option("instrumentation_level") >= 1:
        from dune.perftool.generation import post_include
        post_include("HP_DECLARE_TIMER(driver);\n", filetag="driver")
        contents.insert(0, Line(text="HP_TIMER_START(driver);\n"))
        contents.insert(len(contents) - 1, Line(text="HP_TIMER_STOP(driver);\n"))
        contents.insert(len(contents) - 1, Line(text="DUMP_TIMER({}, driver, {}, true);\n".format(get_option("instrumentation_level"), timestream)))
    contents.insert(0, Line(text="\n"))
    driver_body = Block(contents)
    driver = FunctionBody(driver_signature, driver_body)

    filename = get_option("driver_file")

    from dune.perftool.file import generate_file
    generate_file(filename, "driver", [driver])

    # Reset the caching data structure
    from dune.perftool.generation import delete_cache_items
    delete_cache_items()
